#ifndef HEADER_KRPC_SERVICES_INFERNALROBOTICS
#define HEADER_KRPC_SERVICES_INFERNALROBOTICS

#include <krpc/service.hpp>
#include <krpc/encoder.hpp>
#include <krpc/decoder.hpp>
#include <krpc/stream.hpp>

namespace krpc {
  namespace services {

    class InfernalRobotics : public Service {
    public:
      InfernalRobotics(Client* client);

      // Class forward declarations
      class ControlGroup;
      class Servo;

      /**
       * Returns the servo group with the given name or null if none
       * exists. If multiple servo groups have the same name, only one of them is returned.
       * @param name Name of servo group to find.
       */
      InfernalRobotics::ControlGroup servo_group_with_name(std::string name);

      /**
       * Returns the servo with the given name, from all servo groups, or
       * null if none exists. If multiple servos have the same name, only one of them is returned.
       * @param name Name of the servo to find.
       */
      InfernalRobotics::Servo servo_with_name(std::string name);

      /**
       * A list of all the servo groups in the active vessel.
       */
      std::vector<InfernalRobotics::ControlGroup> servo_groups();

      ::krpc::Stream<InfernalRobotics::ControlGroup> servo_group_with_name_stream(std::string name);

      ::krpc::Stream<InfernalRobotics::Servo> servo_with_name_stream(std::string name);

      ::krpc::Stream<std::vector<InfernalRobotics::ControlGroup>> servo_groups_stream();

      /**
       * A group of servos, obtained by calling InfernalRobotics::servo_groups
       * or InfernalRobotics::servo_group_with_name. Represents the "Servo Groups"
       * in the InfernalRobotics UI.
       */
      class ControlGroup : public krpc::Object<ControlGroup> {
      public:
        ControlGroup(Client* client = NULL, google::protobuf::uint64 id = 0);

        /**
         * Moves all of the servos in the group to the center.
         */
        void move_center();

        /**
         * Moves all of the servos in the group to the left.
         */
        void move_left();

        /**
         * Moves all of the servos in the group to the next preset.
         */
        void move_next_preset();

        /**
         * Moves all of the servos in the group to the previous preset.
         */
        void move_prev_preset();

        /**
         * Moves all of the servos in the group to the right.
         */
        void move_right();

        /**
         * Returns the servo with the given name from this group,
         * or null if none exists.
         * @param name Name of servo to find.
         */
        InfernalRobotics::Servo servo_with_name(std::string name);

        /**
         * Stops the servos in the group.
         */
        void stop();

        /**
         * The key assigned to be the "reverse" key for the group.
         */
        void set_reverse_key(std::string value);

        /**
         * The speed multiplier for the group.
         */
        void set_speed(float value);

        /**
         * The name of the group.
         */
        void set_name(std::string value);

        /**
         * The name of the group.
         */
        std::string name();

        /**
         * The speed multiplier for the group.
         */
        float speed();

        /**
         * Whether the group is expanded in the InfernalRobotics UI.
         */
        bool expanded();

        /**
         * Whether the group is expanded in the InfernalRobotics UI.
         */
        void set_expanded(bool value);

        /**
         * The key assigned to be the "forward" key for the group.
         */
        std::string forward_key();

        /**
         * The key assigned to be the "forward" key for the group.
         */
        void set_forward_key(std::string value);

        /**
         * The servos that are in the group.
         */
        std::vector<InfernalRobotics::Servo> servos();

        /**
         * The key assigned to be the "reverse" key for the group.
         */
        std::string reverse_key();

        ::krpc::Stream<InfernalRobotics::Servo> servo_with_name_stream(std::string name);

        ::krpc::Stream<std::string> name_stream();

        ::krpc::Stream<float> speed_stream();

        ::krpc::Stream<bool> expanded_stream();

        ::krpc::Stream<std::string> forward_key_stream();

        ::krpc::Stream<std::vector<InfernalRobotics::Servo>> servos_stream();

        ::krpc::Stream<std::string> reverse_key_stream();

      };

      /**
       * Represents a servo. Obtained using
       * InfernalRobotics::ControlGroup::servos,
       * InfernalRobotics::ControlGroup::servo_with_name
       * or InfernalRobotics::servo_with_name.
       */
      class Servo : public krpc::Object<Servo> {
      public:
        Servo(Client* client = NULL, google::protobuf::uint64 id = 0);

        /**
         * Moves the servo to the center.
         */
        void move_center();

        /**
         * Moves the servo to the left.
         */
        void move_left();

        /**
         * Moves the servo to the next preset.
         */
        void move_next_preset();

        /**
         * Moves the servo to the previous preset.
         */
        void move_prev_preset();

        /**
         * Moves the servo to the right.
         */
        void move_right();

        /**
         * Moves the servo to position and sets the
         * speed multiplier to speed.
         * @param position The position to move the servo to.
         * @param speed Speed multiplier for the movement.
         */
        void move_to(float position, float speed);

        /**
         * Stops the servo.
         */
        void stop();

        /**
         * Whether the servo is locked.
         */
        void set_is_locked(bool value);

        /**
         * The current speed multiplier set in the UI.
         */
        void set_acceleration(float value);

        /**
         * The current speed at which the servo is moving.
         */
        float current_speed();

        /**
         * The speed multiplier of the servo, specified by the in-game tweak menu.
         */
        float speed();

        /**
         * Whether the servos axis is inverted.
         */
        void set_is_axis_inverted(bool value);

        /**
         * The current speed multiplier set in the UI.
         */
        float acceleration();

        /**
         * The speed multiplier of the servo, specified by the part configuration.
         */
        float config_speed();

        /**
         * Whether the servo is freely moving.
         */
        bool is_free_moving();

        /**
         * The maximum position of the servo, specified by the part configuration.
         */
        float max_config_position();

        /**
         * The maximum position of the servo, specified by the in-game tweak menu.
         */
        float max_position();

        /**
         * The minimum position of the servo, specified by the part configuration.
         */
        float min_config_position();

        /**
         * The maximum position of the servo, specified by the in-game tweak menu.
         */
        void set_max_position(float value);

        /**
         * The speed multiplier of the servo, specified by the in-game tweak menu.
         */
        void set_speed(float value);

        /**
         * The name of the servo.
         */
        void set_name(std::string value);

        /**
         * Whether the servo is moving.
         */
        bool is_moving();

        /**
         * The name of the servo.
         */
        std::string name();

        /**
         * The minimum position of the servo, specified by the in-game tweak menu.
         */
        float min_position();

        /**
         * Whether the servo is locked.
         */
        bool is_locked();

        /**
         * The minimum position of the servo, specified by the in-game tweak menu.
         */
        void set_min_position(float value);

        /**
         * The current speed at which the servo is moving.
         */
        void set_current_speed(float value);

        /**
         * Whether the servo should be highlighted in-game.
         */
        void set_highlight(bool value);

        /**
         * The position of the servo.
         */
        float position();

        /**
         * Whether the servos axis is inverted.
         */
        bool is_axis_inverted();

        ::krpc::Stream<float> current_speed_stream();

        ::krpc::Stream<float> speed_stream();

        ::krpc::Stream<float> acceleration_stream();

        ::krpc::Stream<float> config_speed_stream();

        ::krpc::Stream<bool> is_free_moving_stream();

        ::krpc::Stream<float> max_config_position_stream();

        ::krpc::Stream<float> max_position_stream();

        ::krpc::Stream<float> min_config_position_stream();

        ::krpc::Stream<bool> is_moving_stream();

        ::krpc::Stream<std::string> name_stream();

        ::krpc::Stream<float> min_position_stream();

        ::krpc::Stream<bool> is_locked_stream();

        ::krpc::Stream<float> position_stream();

        ::krpc::Stream<bool> is_axis_inverted_stream();

      };

    };

  }

  namespace services {

    inline InfernalRobotics::InfernalRobotics(Client* client):
      Service(client) {}

    inline InfernalRobotics::ControlGroup InfernalRobotics::servo_group_with_name(std::string name) {
      std::vector<std::string> _args;
      _args.push_back(encoder::encode(name));
      std::string _data = this->_client->invoke("InfernalRobotics", "ServoGroupWithName", _args);
      InfernalRobotics::ControlGroup _result;
      decoder::decode(_result, _data, this->_client);
      return _result;
    }

    inline InfernalRobotics::Servo InfernalRobotics::servo_with_name(std::string name) {
      std::vector<std::string> _args;
      _args.push_back(encoder::encode(name));
      std::string _data = this->_client->invoke("InfernalRobotics", "ServoWithName", _args);
      InfernalRobotics::Servo _result;
      decoder::decode(_result, _data, this->_client);
      return _result;
    }

    inline std::vector<InfernalRobotics::ControlGroup> InfernalRobotics::servo_groups() {
      std::string _data = this->_client->invoke("InfernalRobotics", "get_ServoGroups");
      std::vector<InfernalRobotics::ControlGroup> _result;
      decoder::decode(_result, _data, this->_client);
      return _result;
    }

    inline ::krpc::Stream<InfernalRobotics::ControlGroup> InfernalRobotics::servo_group_with_name_stream(std::string name) {
      std::vector<std::string> _args;
      _args.push_back(encoder::encode(name));
      return ::krpc::Stream<InfernalRobotics::ControlGroup>(this->_client, this->_client->request("InfernalRobotics", "ServoGroupWithName", _args));
    }

    inline ::krpc::Stream<InfernalRobotics::Servo> InfernalRobotics::servo_with_name_stream(std::string name) {
      std::vector<std::string> _args;
      _args.push_back(encoder::encode(name));
      return ::krpc::Stream<InfernalRobotics::Servo>(this->_client, this->_client->request("InfernalRobotics", "ServoWithName", _args));
    }

    inline ::krpc::Stream<std::vector<InfernalRobotics::ControlGroup>> InfernalRobotics::servo_groups_stream() {
      return ::krpc::Stream<std::vector<InfernalRobotics::ControlGroup>>(this->_client, this->_client->request("InfernalRobotics", "get_ServoGroups"));
    }

    inline InfernalRobotics::ControlGroup::ControlGroup(Client* client, google::protobuf::uint64 id):
      Object(client, "InfernalRobotics::ControlGroup", id) {}

    inline void InfernalRobotics::ControlGroup::move_center() {
      std::vector<std::string> _args;
      _args.push_back(encoder::encode(*this));
      this->_client->invoke("InfernalRobotics", "ControlGroup_MoveCenter", _args);
    }

    inline void InfernalRobotics::ControlGroup::move_left() {
      std::vector<std::string> _args;
      _args.push_back(encoder::encode(*this));
      this->_client->invoke("InfernalRobotics", "ControlGroup_MoveLeft", _args);
    }

    inline void InfernalRobotics::ControlGroup::move_next_preset() {
      std::vector<std::string> _args;
      _args.push_back(encoder::encode(*this));
      this->_client->invoke("InfernalRobotics", "ControlGroup_MoveNextPreset", _args);
    }

    inline void InfernalRobotics::ControlGroup::move_prev_preset() {
      std::vector<std::string> _args;
      _args.push_back(encoder::encode(*this));
      this->_client->invoke("InfernalRobotics", "ControlGroup_MovePrevPreset", _args);
    }

    inline void InfernalRobotics::ControlGroup::move_right() {
      std::vector<std::string> _args;
      _args.push_back(encoder::encode(*this));
      this->_client->invoke("InfernalRobotics", "ControlGroup_MoveRight", _args);
    }

    inline InfernalRobotics::Servo InfernalRobotics::ControlGroup::servo_with_name(std::string name) {
      std::vector<std::string> _args;
      _args.push_back(encoder::encode(*this));
      _args.push_back(encoder::encode(name));
      std::string _data = this->_client->invoke("InfernalRobotics", "ControlGroup_ServoWithName", _args);
      InfernalRobotics::Servo _result;
      decoder::decode(_result, _data, this->_client);
      return _result;
    }

    inline void InfernalRobotics::ControlGroup::stop() {
      std::vector<std::string> _args;
      _args.push_back(encoder::encode(*this));
      this->_client->invoke("InfernalRobotics", "ControlGroup_Stop", _args);
    }

    inline void InfernalRobotics::ControlGroup::set_reverse_key(std::string value) {
      std::vector<std::string> _args;
      _args.push_back(encoder::encode(*this));
      _args.push_back(encoder::encode(value));
      this->_client->invoke("InfernalRobotics", "ControlGroup_set_ReverseKey", _args);
    }

    inline void InfernalRobotics::ControlGroup::set_speed(float value) {
      std::vector<std::string> _args;
      _args.push_back(encoder::encode(*this));
      _args.push_back(encoder::encode(value));
      this->_client->invoke("InfernalRobotics", "ControlGroup_set_Speed", _args);
    }

    inline void InfernalRobotics::ControlGroup::set_name(std::string value) {
      std::vector<std::string> _args;
      _args.push_back(encoder::encode(*this));
      _args.push_back(encoder::encode(value));
      this->_client->invoke("InfernalRobotics", "ControlGroup_set_Name", _args);
    }

    inline std::string InfernalRobotics::ControlGroup::name() {
      std::vector<std::string> _args;
      _args.push_back(encoder::encode(*this));
      std::string _data = this->_client->invoke("InfernalRobotics", "ControlGroup_get_Name", _args);
      std::string _result;
      decoder::decode(_result, _data, this->_client);
      return _result;
    }

    inline float InfernalRobotics::ControlGroup::speed() {
      std::vector<std::string> _args;
      _args.push_back(encoder::encode(*this));
      std::string _data = this->_client->invoke("InfernalRobotics", "ControlGroup_get_Speed", _args);
      float _result;
      decoder::decode(_result, _data, this->_client);
      return _result;
    }

    inline bool InfernalRobotics::ControlGroup::expanded() {
      std::vector<std::string> _args;
      _args.push_back(encoder::encode(*this));
      std::string _data = this->_client->invoke("InfernalRobotics", "ControlGroup_get_Expanded", _args);
      bool _result;
      decoder::decode(_result, _data, this->_client);
      return _result;
    }

    inline void InfernalRobotics::ControlGroup::set_expanded(bool value) {
      std::vector<std::string> _args;
      _args.push_back(encoder::encode(*this));
      _args.push_back(encoder::encode(value));
      this->_client->invoke("InfernalRobotics", "ControlGroup_set_Expanded", _args);
    }

    inline std::string InfernalRobotics::ControlGroup::forward_key() {
      std::vector<std::string> _args;
      _args.push_back(encoder::encode(*this));
      std::string _data = this->_client->invoke("InfernalRobotics", "ControlGroup_get_ForwardKey", _args);
      std::string _result;
      decoder::decode(_result, _data, this->_client);
      return _result;
    }

    inline void InfernalRobotics::ControlGroup::set_forward_key(std::string value) {
      std::vector<std::string> _args;
      _args.push_back(encoder::encode(*this));
      _args.push_back(encoder::encode(value));
      this->_client->invoke("InfernalRobotics", "ControlGroup_set_ForwardKey", _args);
    }

    inline std::vector<InfernalRobotics::Servo> InfernalRobotics::ControlGroup::servos() {
      std::vector<std::string> _args;
      _args.push_back(encoder::encode(*this));
      std::string _data = this->_client->invoke("InfernalRobotics", "ControlGroup_get_Servos", _args);
      std::vector<InfernalRobotics::Servo> _result;
      decoder::decode(_result, _data, this->_client);
      return _result;
    }

    inline std::string InfernalRobotics::ControlGroup::reverse_key() {
      std::vector<std::string> _args;
      _args.push_back(encoder::encode(*this));
      std::string _data = this->_client->invoke("InfernalRobotics", "ControlGroup_get_ReverseKey", _args);
      std::string _result;
      decoder::decode(_result, _data, this->_client);
      return _result;
    }

    inline ::krpc::Stream<InfernalRobotics::Servo> InfernalRobotics::ControlGroup::servo_with_name_stream(std::string name) {
      std::vector<std::string> _args;
      _args.push_back(encoder::encode(*this));
      _args.push_back(encoder::encode(name));
      return ::krpc::Stream<InfernalRobotics::Servo>(this->_client, this->_client->request("InfernalRobotics", "ControlGroup_ServoWithName", _args));
    }

    inline ::krpc::Stream<std::string> InfernalRobotics::ControlGroup::name_stream() {
      std::vector<std::string> _args;
      _args.push_back(encoder::encode(*this));
      return ::krpc::Stream<std::string>(this->_client, this->_client->request("InfernalRobotics", "ControlGroup_get_Name", _args));
    }

    inline ::krpc::Stream<float> InfernalRobotics::ControlGroup::speed_stream() {
      std::vector<std::string> _args;
      _args.push_back(encoder::encode(*this));
      return ::krpc::Stream<float>(this->_client, this->_client->request("InfernalRobotics", "ControlGroup_get_Speed", _args));
    }

    inline ::krpc::Stream<bool> InfernalRobotics::ControlGroup::expanded_stream() {
      std::vector<std::string> _args;
      _args.push_back(encoder::encode(*this));
      return ::krpc::Stream<bool>(this->_client, this->_client->request("InfernalRobotics", "ControlGroup_get_Expanded", _args));
    }

    inline ::krpc::Stream<std::string> InfernalRobotics::ControlGroup::forward_key_stream() {
      std::vector<std::string> _args;
      _args.push_back(encoder::encode(*this));
      return ::krpc::Stream<std::string>(this->_client, this->_client->request("InfernalRobotics", "ControlGroup_get_ForwardKey", _args));
    }

    inline ::krpc::Stream<std::vector<InfernalRobotics::Servo>> InfernalRobotics::ControlGroup::servos_stream() {
      std::vector<std::string> _args;
      _args.push_back(encoder::encode(*this));
      return ::krpc::Stream<std::vector<InfernalRobotics::Servo>>(this->_client, this->_client->request("InfernalRobotics", "ControlGroup_get_Servos", _args));
    }

    inline ::krpc::Stream<std::string> InfernalRobotics::ControlGroup::reverse_key_stream() {
      std::vector<std::string> _args;
      _args.push_back(encoder::encode(*this));
      return ::krpc::Stream<std::string>(this->_client, this->_client->request("InfernalRobotics", "ControlGroup_get_ReverseKey", _args));
    }

    inline InfernalRobotics::Servo::Servo(Client* client, google::protobuf::uint64 id):
      Object(client, "InfernalRobotics::Servo", id) {}

    inline void InfernalRobotics::Servo::move_center() {
      std::vector<std::string> _args;
      _args.push_back(encoder::encode(*this));
      this->_client->invoke("InfernalRobotics", "Servo_MoveCenter", _args);
    }

    inline void InfernalRobotics::Servo::move_left() {
      std::vector<std::string> _args;
      _args.push_back(encoder::encode(*this));
      this->_client->invoke("InfernalRobotics", "Servo_MoveLeft", _args);
    }

    inline void InfernalRobotics::Servo::move_next_preset() {
      std::vector<std::string> _args;
      _args.push_back(encoder::encode(*this));
      this->_client->invoke("InfernalRobotics", "Servo_MoveNextPreset", _args);
    }

    inline void InfernalRobotics::Servo::move_prev_preset() {
      std::vector<std::string> _args;
      _args.push_back(encoder::encode(*this));
      this->_client->invoke("InfernalRobotics", "Servo_MovePrevPreset", _args);
    }

    inline void InfernalRobotics::Servo::move_right() {
      std::vector<std::string> _args;
      _args.push_back(encoder::encode(*this));
      this->_client->invoke("InfernalRobotics", "Servo_MoveRight", _args);
    }

    inline void InfernalRobotics::Servo::move_to(float position, float speed) {
      std::vector<std::string> _args;
      _args.push_back(encoder::encode(*this));
      _args.push_back(encoder::encode(position));
      _args.push_back(encoder::encode(speed));
      this->_client->invoke("InfernalRobotics", "Servo_MoveTo", _args);
    }

    inline void InfernalRobotics::Servo::stop() {
      std::vector<std::string> _args;
      _args.push_back(encoder::encode(*this));
      this->_client->invoke("InfernalRobotics", "Servo_Stop", _args);
    }

    inline void InfernalRobotics::Servo::set_is_locked(bool value) {
      std::vector<std::string> _args;
      _args.push_back(encoder::encode(*this));
      _args.push_back(encoder::encode(value));
      this->_client->invoke("InfernalRobotics", "Servo_set_IsLocked", _args);
    }

    inline void InfernalRobotics::Servo::set_acceleration(float value) {
      std::vector<std::string> _args;
      _args.push_back(encoder::encode(*this));
      _args.push_back(encoder::encode(value));
      this->_client->invoke("InfernalRobotics", "Servo_set_Acceleration", _args);
    }

    inline float InfernalRobotics::Servo::current_speed() {
      std::vector<std::string> _args;
      _args.push_back(encoder::encode(*this));
      std::string _data = this->_client->invoke("InfernalRobotics", "Servo_get_CurrentSpeed", _args);
      float _result;
      decoder::decode(_result, _data, this->_client);
      return _result;
    }

    inline float InfernalRobotics::Servo::speed() {
      std::vector<std::string> _args;
      _args.push_back(encoder::encode(*this));
      std::string _data = this->_client->invoke("InfernalRobotics", "Servo_get_Speed", _args);
      float _result;
      decoder::decode(_result, _data, this->_client);
      return _result;
    }

    inline void InfernalRobotics::Servo::set_is_axis_inverted(bool value) {
      std::vector<std::string> _args;
      _args.push_back(encoder::encode(*this));
      _args.push_back(encoder::encode(value));
      this->_client->invoke("InfernalRobotics", "Servo_set_IsAxisInverted", _args);
    }

    inline float InfernalRobotics::Servo::acceleration() {
      std::vector<std::string> _args;
      _args.push_back(encoder::encode(*this));
      std::string _data = this->_client->invoke("InfernalRobotics", "Servo_get_Acceleration", _args);
      float _result;
      decoder::decode(_result, _data, this->_client);
      return _result;
    }

    inline float InfernalRobotics::Servo::config_speed() {
      std::vector<std::string> _args;
      _args.push_back(encoder::encode(*this));
      std::string _data = this->_client->invoke("InfernalRobotics", "Servo_get_ConfigSpeed", _args);
      float _result;
      decoder::decode(_result, _data, this->_client);
      return _result;
    }

    inline bool InfernalRobotics::Servo::is_free_moving() {
      std::vector<std::string> _args;
      _args.push_back(encoder::encode(*this));
      std::string _data = this->_client->invoke("InfernalRobotics", "Servo_get_IsFreeMoving", _args);
      bool _result;
      decoder::decode(_result, _data, this->_client);
      return _result;
    }

    inline float InfernalRobotics::Servo::max_config_position() {
      std::vector<std::string> _args;
      _args.push_back(encoder::encode(*this));
      std::string _data = this->_client->invoke("InfernalRobotics", "Servo_get_MaxConfigPosition", _args);
      float _result;
      decoder::decode(_result, _data, this->_client);
      return _result;
    }

    inline float InfernalRobotics::Servo::max_position() {
      std::vector<std::string> _args;
      _args.push_back(encoder::encode(*this));
      std::string _data = this->_client->invoke("InfernalRobotics", "Servo_get_MaxPosition", _args);
      float _result;
      decoder::decode(_result, _data, this->_client);
      return _result;
    }

    inline float InfernalRobotics::Servo::min_config_position() {
      std::vector<std::string> _args;
      _args.push_back(encoder::encode(*this));
      std::string _data = this->_client->invoke("InfernalRobotics", "Servo_get_MinConfigPosition", _args);
      float _result;
      decoder::decode(_result, _data, this->_client);
      return _result;
    }

    inline void InfernalRobotics::Servo::set_max_position(float value) {
      std::vector<std::string> _args;
      _args.push_back(encoder::encode(*this));
      _args.push_back(encoder::encode(value));
      this->_client->invoke("InfernalRobotics", "Servo_set_MaxPosition", _args);
    }

    inline void InfernalRobotics::Servo::set_speed(float value) {
      std::vector<std::string> _args;
      _args.push_back(encoder::encode(*this));
      _args.push_back(encoder::encode(value));
      this->_client->invoke("InfernalRobotics", "Servo_set_Speed", _args);
    }

    inline void InfernalRobotics::Servo::set_name(std::string value) {
      std::vector<std::string> _args;
      _args.push_back(encoder::encode(*this));
      _args.push_back(encoder::encode(value));
      this->_client->invoke("InfernalRobotics", "Servo_set_Name", _args);
    }

    inline bool InfernalRobotics::Servo::is_moving() {
      std::vector<std::string> _args;
      _args.push_back(encoder::encode(*this));
      std::string _data = this->_client->invoke("InfernalRobotics", "Servo_get_IsMoving", _args);
      bool _result;
      decoder::decode(_result, _data, this->_client);
      return _result;
    }

    inline std::string InfernalRobotics::Servo::name() {
      std::vector<std::string> _args;
      _args.push_back(encoder::encode(*this));
      std::string _data = this->_client->invoke("InfernalRobotics", "Servo_get_Name", _args);
      std::string _result;
      decoder::decode(_result, _data, this->_client);
      return _result;
    }

    inline float InfernalRobotics::Servo::min_position() {
      std::vector<std::string> _args;
      _args.push_back(encoder::encode(*this));
      std::string _data = this->_client->invoke("InfernalRobotics", "Servo_get_MinPosition", _args);
      float _result;
      decoder::decode(_result, _data, this->_client);
      return _result;
    }

    inline bool InfernalRobotics::Servo::is_locked() {
      std::vector<std::string> _args;
      _args.push_back(encoder::encode(*this));
      std::string _data = this->_client->invoke("InfernalRobotics", "Servo_get_IsLocked", _args);
      bool _result;
      decoder::decode(_result, _data, this->_client);
      return _result;
    }

    inline void InfernalRobotics::Servo::set_min_position(float value) {
      std::vector<std::string> _args;
      _args.push_back(encoder::encode(*this));
      _args.push_back(encoder::encode(value));
      this->_client->invoke("InfernalRobotics", "Servo_set_MinPosition", _args);
    }

    inline void InfernalRobotics::Servo::set_current_speed(float value) {
      std::vector<std::string> _args;
      _args.push_back(encoder::encode(*this));
      _args.push_back(encoder::encode(value));
      this->_client->invoke("InfernalRobotics", "Servo_set_CurrentSpeed", _args);
    }

    inline void InfernalRobotics::Servo::set_highlight(bool value) {
      std::vector<std::string> _args;
      _args.push_back(encoder::encode(*this));
      _args.push_back(encoder::encode(value));
      this->_client->invoke("InfernalRobotics", "Servo_set_Highlight", _args);
    }

    inline float InfernalRobotics::Servo::position() {
      std::vector<std::string> _args;
      _args.push_back(encoder::encode(*this));
      std::string _data = this->_client->invoke("InfernalRobotics", "Servo_get_Position", _args);
      float _result;
      decoder::decode(_result, _data, this->_client);
      return _result;
    }

    inline bool InfernalRobotics::Servo::is_axis_inverted() {
      std::vector<std::string> _args;
      _args.push_back(encoder::encode(*this));
      std::string _data = this->_client->invoke("InfernalRobotics", "Servo_get_IsAxisInverted", _args);
      bool _result;
      decoder::decode(_result, _data, this->_client);
      return _result;
    }

    inline ::krpc::Stream<float> InfernalRobotics::Servo::current_speed_stream() {
      std::vector<std::string> _args;
      _args.push_back(encoder::encode(*this));
      return ::krpc::Stream<float>(this->_client, this->_client->request("InfernalRobotics", "Servo_get_CurrentSpeed", _args));
    }

    inline ::krpc::Stream<float> InfernalRobotics::Servo::speed_stream() {
      std::vector<std::string> _args;
      _args.push_back(encoder::encode(*this));
      return ::krpc::Stream<float>(this->_client, this->_client->request("InfernalRobotics", "Servo_get_Speed", _args));
    }

    inline ::krpc::Stream<float> InfernalRobotics::Servo::acceleration_stream() {
      std::vector<std::string> _args;
      _args.push_back(encoder::encode(*this));
      return ::krpc::Stream<float>(this->_client, this->_client->request("InfernalRobotics", "Servo_get_Acceleration", _args));
    }

    inline ::krpc::Stream<float> InfernalRobotics::Servo::config_speed_stream() {
      std::vector<std::string> _args;
      _args.push_back(encoder::encode(*this));
      return ::krpc::Stream<float>(this->_client, this->_client->request("InfernalRobotics", "Servo_get_ConfigSpeed", _args));
    }

    inline ::krpc::Stream<bool> InfernalRobotics::Servo::is_free_moving_stream() {
      std::vector<std::string> _args;
      _args.push_back(encoder::encode(*this));
      return ::krpc::Stream<bool>(this->_client, this->_client->request("InfernalRobotics", "Servo_get_IsFreeMoving", _args));
    }

    inline ::krpc::Stream<float> InfernalRobotics::Servo::max_config_position_stream() {
      std::vector<std::string> _args;
      _args.push_back(encoder::encode(*this));
      return ::krpc::Stream<float>(this->_client, this->_client->request("InfernalRobotics", "Servo_get_MaxConfigPosition", _args));
    }

    inline ::krpc::Stream<float> InfernalRobotics::Servo::max_position_stream() {
      std::vector<std::string> _args;
      _args.push_back(encoder::encode(*this));
      return ::krpc::Stream<float>(this->_client, this->_client->request("InfernalRobotics", "Servo_get_MaxPosition", _args));
    }

    inline ::krpc::Stream<float> InfernalRobotics::Servo::min_config_position_stream() {
      std::vector<std::string> _args;
      _args.push_back(encoder::encode(*this));
      return ::krpc::Stream<float>(this->_client, this->_client->request("InfernalRobotics", "Servo_get_MinConfigPosition", _args));
    }

    inline ::krpc::Stream<bool> InfernalRobotics::Servo::is_moving_stream() {
      std::vector<std::string> _args;
      _args.push_back(encoder::encode(*this));
      return ::krpc::Stream<bool>(this->_client, this->_client->request("InfernalRobotics", "Servo_get_IsMoving", _args));
    }

    inline ::krpc::Stream<std::string> InfernalRobotics::Servo::name_stream() {
      std::vector<std::string> _args;
      _args.push_back(encoder::encode(*this));
      return ::krpc::Stream<std::string>(this->_client, this->_client->request("InfernalRobotics", "Servo_get_Name", _args));
    }

    inline ::krpc::Stream<float> InfernalRobotics::Servo::min_position_stream() {
      std::vector<std::string> _args;
      _args.push_back(encoder::encode(*this));
      return ::krpc::Stream<float>(this->_client, this->_client->request("InfernalRobotics", "Servo_get_MinPosition", _args));
    }

    inline ::krpc::Stream<bool> InfernalRobotics::Servo::is_locked_stream() {
      std::vector<std::string> _args;
      _args.push_back(encoder::encode(*this));
      return ::krpc::Stream<bool>(this->_client, this->_client->request("InfernalRobotics", "Servo_get_IsLocked", _args));
    }

    inline ::krpc::Stream<float> InfernalRobotics::Servo::position_stream() {
      std::vector<std::string> _args;
      _args.push_back(encoder::encode(*this));
      return ::krpc::Stream<float>(this->_client, this->_client->request("InfernalRobotics", "Servo_get_Position", _args));
    }

    inline ::krpc::Stream<bool> InfernalRobotics::Servo::is_axis_inverted_stream() {
      std::vector<std::string> _args;
      _args.push_back(encoder::encode(*this));
      return ::krpc::Stream<bool>(this->_client, this->_client->request("InfernalRobotics", "Servo_get_IsAxisInverted", _args));
    }

  }

}

#endif
